﻿Imports Microsoft.VisualBasic
Imports System
Imports System.Text

    ' The Porter2 aka SnowBall stemming algorithm 
    ' 
    ' VB.NET port from C# by Hunter Beanland - http://beanland.net.au
    ' v1.0.0, 2010-02-11
    ' Sections have been ported via compilation and disassembly of MSIL to VB, 
    ' therefore may appear to be compiler optimised in code (strange coding)
    '
    ' C# implementation written by Kamil Bartocha, email: crow2 (at) poczta.fm
    ' C# version: 1.0.0 - date: 11.04.2007
    ' 
    ' This is a VB.NET implementation of Martin Porter's "Porter2" aka "SnowBall" algorithm as described at:
    ' http://snowball.tartarus.org/algorithms/english/stemmer.html
    ' This algorithm does not always produce a correctly spelled word. ie "Services" becomes "Servic" (typically "e" suffixes are an issue)
    ' as English does not strictly follow rules. This is very good, but the best we can hope for without using a dictionary lookup.
    ' It is good enough for searching or matching purposes - which is probably what you want to do.
    ' 
    ' It is provided as a ready-to-use class with only a single public method:
    ' public function stem(word as string) as string
    ' 
    ' Usage example:
    ' 
    ' Dim porterStemmer As New Stemmer()
    ' Dim output As String = porterStemmer.stem("word")
    ' 
    ' 
    ' Notices for users: 
    ' 
    ' 1. The sb.Replace(...) code might be somehow optimized but this
    ' seemed to be the best choice to do it.
    ' 
    ' 2. I've tried to use the String class more often insted of the sb.ToString(...)
    ' but the resulting code was 100% SLOWER! I guess it's because of StringBuilder's 
    ' internal optimizations.
    '
    ' 
    ' LICENSE:
    ' 
    ' Copyright 2007 Kamil Bartocha. All rights reserved.
    ' 
    ' Redistribution and use in source and binary forms, 
    ' with or without modification, are permitted provided that 
    ' the following conditions are met:
    ' 
    ' 1. Redistributions of source code must retain the above copyright notice, 
    ' this list of conditions and the following disclaimer.
    ' 
    ' 2. Redistributions in binary form must reproduce the above copyright notice, 
    ' this list of conditions and the following disclaimer in the documentation 
    ' and/or other materials provided with the distribution.
    ' 
    ' THIS SOFTWARE IS PROVIDED BY KAMIL BARTOCHA ``AS IS'' 
    ' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
    ' THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
    ' PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE FREEBSD PROJECT OR CONTRIBUTORS 
    ' BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
    ' OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
    ' OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
    ' OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
    ' WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
    ' ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
    ' EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    Public Class Stemmer

        Private doubles() As String = {"bb", "dd", "ff", "gg", "mm", "nn", "pp", "rr", "tt"}
        Private validLiEndings() As String = {"c", "d", "e", "g", "h", "k", "m", "n", "r", "t"}
        Private step1bReplacements(,) As String = {{"eedly", "ee"}, {"ingly", ""}, {"edly", ""}, {"eed", "ee"}, {"ing", ""}, {"ed", ""}}
        Private step2Replacements(,) As String = {{"ization", "ize"}, {"iveness", "ive"}, {"fulness", "ful"}, {"ational", "ate"}, {"ousness", "ous"}, {"biliti", "ble"}, {"tional", "tion"}, {"lessli", "less"}, {"fulli", "ful"}, {"entli", "ent"}, {"ation", "ate"}, {"aliti", "al"}, {"iviti", "ive"}, {"ousli", "ous"}, {"alism", "al"}, {"abli", "able"}, {"anci", "ance"}, {"alli", "al"}, {"izer", "ize"}, {"enci", "ence"}, {"ator", "ate"}, {"bli", "ble"}, {"ogi", "og"}, {"li", ""}}
        Private step3Replacements(,) As String = {{"ational", "ate"}, {"tional", "tion"}, {"alize", "al"}, {"icate", "ic"}, {"iciti", "ic"}, {"ative", ""}, {"ical", "ic"}, {"ness", ""}, {"ful", ""}}
        Private step4Replacements() As String = {"ement", "ment", "able", "ible", "ance", "ence", "ate", "iti", "ion", "ize", "ive", "ous", "ant", "ism", "ent", "al", "er", "ic"}
        Private exceptions(,) As String = {{"skis", "ski"}, {"skies", "sky"}, {"dying", "die"}, {"lying", "lie"}, {"tying", "tie"}, {"idly", "idl"}, {"gently", "gentl"}, {"ugly", "ugli"}, {"early", "earli"}, {"only", "onli"}, {"singly", "singl"}, {"sky", "sky"}, {"news", "news"}, {"howe", "howe"}, {"atlas", "atlas"}, {"cosmos", "cosmos"}, {"bias", "bias"}, {"andes", "andes"}}
        Private exceptions2() As String = {"inning", "outing", "canning", "herring", "earring", "proceed", "exceed", "succeed"}

        'A helper table lookup code - used for vowel lookup 
        Private Function arrayContains(ByVal arr As String(), ByVal s As String) As Boolean
            For i As Integer = 0 To arr.Length - 1
                If (arr(i) = s) Then Return True
            Next i
            Return False
        End Function

        Private Function isVowel(ByVal s As StringBuilder, ByVal offset As Integer) As Boolean
            Select Case s.Chars(offset)
                Case "o"c, "u"c, "y"c, "a"c, "e"c, "i"c
                    Return True
            End Select
            Return False
        End Function

        Private Function isShortSyllable(ByVal s As StringBuilder, ByVal offset As Integer) As Boolean
            Return (Not (((offset <> 0) OrElse Not isVowel(s, 0)) OrElse isVowel(s, 1)) OrElse Not (((((offset <= 0) OrElse (offset >= (s.Length - 1))) OrElse (Not isVowel(s, offset) OrElse isVowel(s, (offset + 1)))) OrElse (((s.Chars((offset + 1)) = "w"c) OrElse (s.Chars((offset + 1)) = "x"c)) OrElse (s.Chars((offset + 1)) = "Y"c))) OrElse isVowel(s, (offset - 1))))
        End Function

        Private Function isShortWord(ByVal s As StringBuilder, ByVal r1 As Integer) As Boolean
            Return ((r1 >= s.Length) AndAlso isShortSyllable(s, (s.Length - 2)))
        End Function

        Private Sub changeY(ByVal sb As StringBuilder)
            If (sb.Chars(0) = "y"c) Then
                sb.Chars(0) = "Y"c
            End If
            For i As Integer = 1 To sb.Length - 1
                If ((sb.Chars(i) = "y"c) AndAlso isVowel(sb, (i - 1))) Then
                    sb.Chars(i) = "Y"c
                End If
            Next i
        End Sub

        Private Sub computeR1R2(ByVal sb As StringBuilder, ByRef r1 As Integer, ByRef r2 As Integer)
            Dim i As Integer
            r1 = sb.Length
            r2 = sb.Length
            If ((sb.Length >= 5) AndAlso ((sb.ToString(0, 5) = "gener") OrElse (sb.ToString(0, 5) = "arsen"))) Then
                r1 = 5
            End If
            If ((sb.Length >= 6) AndAlso (sb.ToString(0, 6) = "commun")) Then
                r1 = 6
            End If
            If (r1 = sb.Length) Then 'If R1 has not been changed by exception words
                For i = 1 To sb.Length - 1 'Compute R1 according to the algorithm
                    If Not (isVowel(sb, i) OrElse Not isVowel(sb, (i - 1))) Then
                        r1 = (i + 1)
                        Exit For
                    End If
                Next i
            End If

            For i = (r1 + 1) To sb.Length - 1
                If Not (isVowel(sb, i) OrElse Not isVowel(sb, (i - 1))) Then
                    r2 = (i + 1)
                    Exit For
                End If
            Next i
        End Sub

        Private Sub step0(ByVal sb As StringBuilder)
            If ((sb.Length >= 3) AndAlso (sb.ToString((sb.Length - 3), 3) = "'s'")) Then
                sb.Remove((sb.Length - 3), 3)
            ElseIf ((sb.Length >= 2) AndAlso (sb.ToString((sb.Length - 2), 2) = "'s")) Then
                sb.Remove((sb.Length - 2), 2)
            ElseIf (sb.Chars((sb.Length - 1)) = "'"c) Then
                sb.Remove((sb.Length - 1), 1)
            End If
        End Sub

        Private Sub step1a(ByVal sb As StringBuilder)
            If ((sb.Length >= 4) AndAlso (sb.ToString((sb.Length - 4), 4) = "sses")) Then
                sb.Replace("sses", "ss", (sb.Length - 4), 4)
            ElseIf ((sb.Length >= 3) AndAlso ((sb.ToString((sb.Length - 3), 3) = "ied") OrElse (sb.ToString((sb.Length - 3), 3) = "ies"))) Then
                If (sb.Length > 4) Then
                    sb.Replace(sb.ToString((sb.Length - 3), 3), "i", (sb.Length - 3), 3)
                Else
                    sb.Replace(sb.ToString((sb.Length - 3), 3), "ie", (sb.Length - 3), 3)
                End If
            ElseIf (((sb.Length < 2) OrElse ((sb.ToString((sb.Length - 2), 2) <> "us") AndAlso Not (sb.ToString((sb.Length - 2), 2) = "ss"))) AndAlso ((sb.Length > 0) AndAlso (sb.ToString((sb.Length - 1), 1) = "s"))) Then
                Dim i As Integer
                For i = 0 To (sb.Length - 2) - 1
                    If isVowel(sb, i) Then
                        sb.Remove((sb.Length - 1), 1)
                        Exit For
                    End If
                Next i
            End If
        End Sub

        Private Sub step1b(ByVal sb As StringBuilder, ByVal r1 As Integer)
            Dim i As Integer
            For i = 0 To 6 - 1 Step +1
                If (sb.Length > step1bReplacements(i, 0).Length) AndAlso (sb.ToString(sb.Length - step1bReplacements(i, 0).Length, step1bReplacements(i, 0).Length) = step1bReplacements(i, 0)) Then
                    Select Case step1bReplacements(i, 0)
                        Case "eedly", "eed"
                            If sb.Length - step1bReplacements(i, 0).Length >= r1 Then
                                sb.Replace(step1bReplacements(i, 0), step1bReplacements(i, 1), sb.Length - step1bReplacements(i, 0).Length, step1bReplacements(i, 0).Length)
                            End If
                            Exit For
                        Case Else
                            Dim found As Boolean = False
                            Dim j As Integer
                            For j = 0 To sb.Length - step1bReplacements(i, 0).Length - 1 Step +1
                                If isVowel(sb, j) Then
                                    sb.Replace(step1bReplacements(i, 0), step1bReplacements(i, 1), sb.Length - step1bReplacements(i, 0).Length, step1bReplacements(i, 0).Length)
                                    found = True
                                    Exit For
                                End If
                            Next
                            If Not found Then
                                Return
                            End If
                            Select Case sb.ToString(sb.Length - 2, 2)
                                Case "at", "bl", "iz"
                                    sb.Append("e")
                                    Return
                            End Select
                            If arrayContains(doubles, sb.ToString(sb.Length - 2, 2)) Then
                                sb.Remove(sb.Length - 1, 1)
                                Return
                            End If
                            If isShortWord(sb, r1) Then
                                sb.Append("e")
                            End If
                            Exit For
                    End Select
                    Return
                End If
            Next
        End Sub

        Private Sub step1c(ByVal sb As StringBuilder)
            If (sb.Length > 0) AndAlso (sb(sb.Length - 1) = "y"c OrElse sb(sb.Length - 1) = "Y"c) AndAlso (sb.Length > 2) AndAlso (Not isVowel(sb, sb.Length - 2)) Then
                sb(sb.Length - 1) = "i"c
            End If
        End Sub

        Private Sub step2(ByVal sb As StringBuilder, ByVal r1 As Integer)
            For i As Integer = 0 To 23
                If (sb.Length >= step2Replacements(i, 0).Length) AndAlso (sb.ToString(sb.Length - step2Replacements(i, 0).Length, step2Replacements(i, 0).Length) = step2Replacements(i, 0)) Then
                    If sb.Length - step2Replacements(i, 0).Length >= r1 Then
                        Select Case step2Replacements(i, 0)
                            Case "ogi"
                                If (sb.Length > 3) AndAlso (sb(sb.Length - step2Replacements(i, 0).Length - 1) = "l"c) Then
                                    sb.Replace(step2Replacements(i, 0), step2Replacements(i, 1), sb.Length - step2Replacements(i, 0).Length, step2Replacements(i, 0).Length)
                                End If
                                Return
                            Case "li"
                                If (sb.Length > 1) AndAlso (arrayContains(validLiEndings, sb.ToString(sb.Length - 3, 1))) Then
                                    sb.Remove(sb.Length - 2, 2)
                                End If
                                Return
                            Case Else
                                sb.Replace(step2Replacements(i, 0), step2Replacements(i, 1), sb.Length - step2Replacements(i, 0).Length, step2Replacements(i, 0).Length)
                                Return
                                Exit Select
                        End Select
                    Else
                        Return
                    End If
                End If
            Next
        End Sub

        Private Sub step3(ByVal sb As StringBuilder, ByVal r1 As Integer, ByVal r2 As Integer)
            Dim i As Integer
            For i = 0 To 9 - 1
                If ((sb.Length >= step3Replacements(i, 0).Length) AndAlso (sb.ToString((sb.Length - step3Replacements(i, 0).Length), step3Replacements(i, 0).Length) = step3Replacements(i, 0))) Then
                    If ((sb.Length - step3Replacements(i, 0).Length) >= r1) Then
                        Dim stepReplacement As String = step3Replacements(i, 0)
                        If ((Not stepReplacement Is Nothing) AndAlso (stepReplacement = "ative")) Then
                            If ((sb.Length - step3Replacements(i, 0).Length) >= r2) Then
                                sb.Replace(step3Replacements(i, 0), step3Replacements(i, 1), (sb.Length - step3Replacements(i, 0).Length), step3Replacements(i, 0).Length)
                            End If
                        Else
                            sb.Replace(step3Replacements(i, 0), step3Replacements(i, 1), (sb.Length - step3Replacements(i, 0).Length), step3Replacements(i, 0).Length)
                        End If
                    End If
                    Exit For
                End If
            Next i
        End Sub

        Private Sub step4(ByVal sb As StringBuilder, ByVal r2 As Integer)
            Dim i As Integer
            For i = 0 To 18 - 1
                If ((sb.Length >= step4Replacements(i).Length) AndAlso (sb.ToString((sb.Length - step4Replacements(i).Length), step4Replacements(i).Length) = step4Replacements(i))) Then
                    If ((sb.Length - step4Replacements(i).Length) >= r2) Then
                        Dim stepReplacement As String = step4Replacements(i)
                        If ((Not stepReplacement Is Nothing) AndAlso (stepReplacement = "ion")) Then
                            If ((sb.Length > 3) AndAlso ((sb.Chars(((sb.Length - step4Replacements(i).Length) - 1)) = "s"c) OrElse (sb.Chars(((sb.Length - step4Replacements(i).Length) - 1)) = "t"c))) Then
                                sb.Remove((sb.Length - step4Replacements(i).Length), step4Replacements(i).Length)
                            End If
                        Else
                            sb.Remove((sb.Length - step4Replacements(i).Length), step4Replacements(i).Length)
                        End If
                    End If
                    Exit For
                End If
            Next i
        End Sub

        Private Sub step5(ByVal sb As StringBuilder, ByVal r1 As Integer, ByVal r2 As Integer)
            If (sb.Length > 0) Then
                If ((sb.Chars((sb.Length - 1)) = "e"c) AndAlso (((sb.Length - 1) >= r2) OrElse (((sb.Length - 1) >= r1) AndAlso Not isShortSyllable(sb, (sb.Length - 3))))) Then
                    sb.Remove((sb.Length - 1), 1)
                ElseIf (((sb.Chars((sb.Length - 1)) = "l"c) AndAlso ((sb.Length - 1) >= r2)) AndAlso (sb.Chars((sb.Length - 2)) = "l"c)) Then
                    sb.Remove((sb.Length - 1), 1)
                End If
            End If
        End Sub

        Public Function stem(ByVal word As String) As String
            Dim i As Integer
            If (word.Length < 3) Then
                Return word
            End If
            Dim sb As New StringBuilder(word.ToLower)
            If (sb.Chars(0) = "'"c) Then
                sb.Remove(0, 1)
            End If
            For i = 0 To (exceptions.Length / 2) - 1
                If (word = exceptions(i, 0)) Then
                    Return exceptions(i, 1)
                End If
            Next i
            Dim r1 As Integer = 0
            Dim r2 As Integer = 0
            changeY(sb)
            computeR1R2(sb, (r1), (r2))
            step0(sb)
            step1a(sb)
            For i = 0 To exceptions2.Length - 1
                If (sb.ToString = exceptions2(i)) Then
                    Return exceptions2(i)
                End If
            Next i
            step1b(sb, r1)
            step1c(sb)
            step2(sb, r1)
            step3(sb, r1, r2)
            step4(sb, r2)
            step5(sb, r1, r2)
            Return sb.ToString.ToLower
        End Function
    End Class

